#include <time.h>
#include <string.h>

#define FXN_TEST        0

#if !FXN_TEST
#define mygmtime        gmtime
#define mymktime        mktime
#define mylocaltime     localtime
#define mytime          time
#endif

/*
  time_t is time in seconds since GMT 00:00:00 1970-Jan-01 (aka "the Epoch")

  struct tm:
  tm_sec        [0,60]
  tm_min        [0,59]
  tm_hour       [0,23]
  tm_mday       [1,31]          day of month
  tm_mon        [0,11]
  tm_year       [70,138]        year since 1900
  tm_wday       [0,6]           day since Sunday (day of week)
  tm_yday       [0,365]         day since January 1 (day of year)
  tm_isdst      [<0,0,>0]       daylight saving time flag

  leap year determination:
  y%400 y%100   y%4
    (!0)  (!0)   !0     not leap
    (!0)   !0     0     leap
     !0     0     0     not leap
      0     0     0     leap

  period is 400 years
  every 400 years contain exactly 97 leap years
  400 years = 146097 (3*3*3*7*773) days
  1 year = 365.2425 days

  signed 32-bit time_t format can fit 68 whole years, yielding [1970,2038)
  DOS FAT file time is 128-whole-years-long, yielding [1980,2108)
*/

static
int IsLeapYear (unsigned year)
{
  if (year & 3)
    return 0;

  year %= 400;

  if (!(year-=100)) return 0;
  if (!(year-=100)) return 0;
  if (!(year-=100)) return 0;

  return 1;
}

static
const unsigned aDaysSinceEpoch[]=
{
            0,  365,  730, 1096, 1461, 1826, 2191,      // 1970, ...
   2557, 2922, 3287, 3652, 4018, 4383, 4748, 5113,
   5479, 5844, 6209, 6574, 6940, 7305, 7670, 8035,
   8401, 8766, 9131, 9496, 9862,10227,10592,10957,
  11323,11688,12053,12418,12784,13149,13514,13879,
  14245,14610,14975,15340,15706,16071,16436,16801,
  17167,17532,17897,18262,18628,18993,19358,19723,
  20089,20454,20819,21184,21550,21915,22280,22645,
  23011,23376,23741,24106,24472,24837,25202             // ..., 2039
};

#if 0
// has been used to generate aDaysSinceEpoch[]:
void mktab()
{
  unsigned y, d=0;
  for (y=1970;y<=2106;d+=365+IsLeapYear(y),y++)
    printf ("%5u,", d), (y%8) ?0:printf ("\n");
  printf ("\n");
}
#endif

static
const unsigned aDaysSinceJan1st[2][13]=
{
  {0,31,59,90,120,151,181,212,243,273,304,334,365},     // 365 days, nonleap
  {0,31,60,91,121,152,182,213,244,274,305,335,366}      // 366 days, leap
};

#define SECONDS_PER_MINUTE      60
#define SECONDS_PER_HOUR        (60*SECONDS_PER_MINUTE)
#define SECONDS_PER_DAY         (24UL*SECONDS_PER_HOUR)

struct tm* mygmtime (const time_t* pTime)
{
  static struct tm BrokenDownTime;
  time_t trem = *pTime;
  unsigned DaysPassed;
  int lidx, ridx, idx, IsLeap;

  // In case of error gmtime() must return a non-NULL pointer to tm struct,
  // so just zeroing the struct and returning its address:
  if ((trem < 0) || (trem == (time_t)-1))
  {
lerr:
    memset (&BrokenDownTime, 0, sizeof(BrokenDownTime));
    return &BrokenDownTime;
  }

  DaysPassed = trem / SECONDS_PER_DAY;
  trem -= DaysPassed * SECONDS_PER_DAY;

  BrokenDownTime.tm_wday = (DaysPassed + 4) % 7;

  BrokenDownTime.tm_hour = trem / SECONDS_PER_HOUR;
  trem -= BrokenDownTime.tm_hour * (time_t)SECONDS_PER_HOUR;

  BrokenDownTime.tm_min = trem / SECONDS_PER_MINUTE;
  trem -= BrokenDownTime.tm_min * SECONDS_PER_MINUTE;

  BrokenDownTime.tm_sec = trem;

  // Convert integer amount of days since 1970-Jan-01 to the year:
  if (DaysPassed >= 
      aDaysSinceEpoch[sizeof(aDaysSinceEpoch)/sizeof(aDaysSinceEpoch[0])-1])
    goto lerr;
  for (lidx=0, ridx=sizeof(aDaysSinceEpoch)/sizeof(aDaysSinceEpoch[0])-1;;)
  {
    idx = (lidx + ridx) >> 1;
    if (DaysPassed < aDaysSinceEpoch[idx])
      ridx = idx;
    else if (DaysPassed >= aDaysSinceEpoch[idx+1])
      lidx = idx;
    else
      break;
  }
  BrokenDownTime.tm_year = 70 + idx;
  IsLeap = IsLeapYear (1970 + idx);
  DaysPassed -= aDaysSinceEpoch[idx];

  BrokenDownTime.tm_yday = DaysPassed;

  // Convert integer amount of days since Jan-01 to the month:
  for (lidx=0, ridx=sizeof(aDaysSinceJan1st[0])/sizeof(aDaysSinceJan1st[0][0])-1;;)
  {
    idx = (lidx + ridx) >> 1;
    if (DaysPassed < aDaysSinceJan1st[IsLeap][idx])
      ridx = idx;
    else if (DaysPassed >= aDaysSinceJan1st[IsLeap][idx+1])
      lidx = idx;
    else
      break;
  }
  BrokenDownTime.tm_mon = idx;
  DaysPassed -= aDaysSinceJan1st[IsLeap][idx];

  BrokenDownTime.tm_mday = 1 + DaysPassed;

  BrokenDownTime.tm_isdst = 0;  // no daylight saving time

  return &BrokenDownTime;
}

#define SECONDS_PER_YEAR        (365UL*SECONDS_PER_DAY)

time_t mymktime (struct tm* pBrokenDownTime)
{
  time_t res;
  int tm_yday, IsLeap, DaysInMonth;

  if
  (
    (pBrokenDownTime->tm_sec < 0) || (pBrokenDownTime->tm_sec >  60) ||
    (pBrokenDownTime->tm_min < 0) || (pBrokenDownTime->tm_min >  59) ||
    (pBrokenDownTime->tm_hour< 0) || (pBrokenDownTime->tm_hour>  23) ||
    (pBrokenDownTime->tm_mday< 1) ||
    (pBrokenDownTime->tm_mon < 0) || (pBrokenDownTime->tm_mon >  11) ||
    (pBrokenDownTime->tm_year<70) || (pBrokenDownTime->tm_year> 138)
  )
    return (time_t)-1;

  IsLeap = IsLeapYear (1900 + pBrokenDownTime->tm_year);

  DaysInMonth = (pBrokenDownTime->tm_mon == 1) ?
                28 + IsLeap:
                30 + ((0xAD5 >> pBrokenDownTime->tm_mon) & 1);
  if (pBrokenDownTime->tm_mday > DaysInMonth)
    return (time_t)-1;

  tm_yday = aDaysSinceJan1st[IsLeap][pBrokenDownTime->tm_mon] +
            pBrokenDownTime->tm_mday - 1;

  res = pBrokenDownTime->tm_sec +
        pBrokenDownTime->tm_min * SECONDS_PER_MINUTE +
        pBrokenDownTime->tm_hour * (time_t)SECONDS_PER_HOUR +
                         tm_yday * SECONDS_PER_DAY +
        (pBrokenDownTime->tm_year - 70) * SECONDS_PER_YEAR;

  // Now account for extra days in the leap years between 1970 and given year:
  //   add a day every 4 years starting in 1973,
  //   subtract a day back out every 100 years starting in 2001, and
  //   add a day back in every 400 years starting in 2001:
  res += ((pBrokenDownTime->tm_year-69)>>2) * (time_t)SECONDS_PER_DAY -
         ((pBrokenDownTime->tm_year-1)/100) * (time_t)SECONDS_PER_DAY +
         ((pBrokenDownTime->tm_year+299)/400) * (time_t)SECONDS_PER_DAY;

  return res;
}

struct tm* mylocaltime (const time_t* pTime)
{
  // for now, no time zones supported,
  // so the local and greenwich times are the same:
  return mygmtime (pTime);
}

#if !FXN_TEST
#ifdef __cplusplus
extern "C" {
#endif
extern void _GetDateAndTime (unsigned* pYear, unsigned* pMonth, unsigned* pDay,
                             unsigned* pHour, unsigned* pMinute, unsigned* pSecond);
#ifdef __cplusplus
}
#endif

time_t mytime (time_t* pTime)
{
  time_t res;
  struct tm BrokenDownTime;
  // for now, no time zones supported,
  // so the local and greenwich times are the same:
  _GetDateAndTime
  (
    (unsigned*)&BrokenDownTime.tm_year,
    (unsigned*)&BrokenDownTime.tm_mon,
    (unsigned*)&BrokenDownTime.tm_mday,
    (unsigned*)&BrokenDownTime.tm_hour,
    (unsigned*)&BrokenDownTime.tm_min,
    (unsigned*)&BrokenDownTime.tm_sec
  );
  BrokenDownTime.tm_mon--;      // tm_mon is 0 through 11
  BrokenDownTime.tm_year-=1900; // tm_year is 1900-based
  BrokenDownTime.tm_isdst = 0;  // no daylight saving time
  res = mktime (&BrokenDownTime);
  if (pTime != NULL)
    *pTime = res;
  return res;
}
#endif

#if FXN_TEST
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>

int main()
{
  time_t t1, t2;
  struct tm *pt1, *pt2;

  srand ((unsigned)time(&t1));
  for(;;)
  {
    if (kbhit()) break;
    t1 = ((time_t)(rand()&0x7FFFu)<<16)|
         ((time_t)(rand()&0x7FFFu)<<1)|(rand()&1);
    pt1 = gmtime (&t1);
    pt2 = mygmtime (&t1);
    printf ("t=%lu\n", (unsigned long)t1);
    printf
    (
      "YYYY-MM-DD HH:MM:SS | DOY DOW F\n"
      "%4u-%02u-%02u %02u:%02u:%02u | %3u %3u %d\n"
      "%4u-%02u-%02u %02u:%02u:%02u | %3u %3u %d\n",

      pt1->tm_year+1900, pt1->tm_mon+1, pt1->tm_mday,
      pt1->tm_hour, pt1->tm_min, pt1->tm_sec,
      pt1->tm_yday, pt1->tm_wday, pt1->tm_isdst,

      pt2->tm_year+1900, pt2->tm_mon+1, pt2->tm_mday,
      pt2->tm_hour, pt2->tm_min, pt2->tm_sec,
      pt2->tm_yday, pt2->tm_wday, pt2->tm_isdst
    );
    if
    (
      (pt1->tm_year != pt2->tm_year) ||
      (pt1->tm_mon  != pt2->tm_mon ) ||
      (pt1->tm_mday != pt2->tm_mday) ||
      (pt1->tm_hour != pt2->tm_hour) ||
      (pt1->tm_min  != pt2->tm_min ) ||
      (pt1->tm_sec  != pt2->tm_sec ) ||
      (pt1->tm_yday != pt2->tm_yday) ||
      (pt1->tm_wday != pt2->tm_wday)
    )
    {
      printf ("*gmtime() error\n");
      break;
    }
    t2 = mymktime (pt2);
    if (t2 != t1)
    {
      printf ("*mktime() error\n");
      break;
    }
    printf ("\n");
  }

  return 0;
}
#endif
